/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 * e-passwords.c
 *
 * Copyright (C) 2001 Ximian, Inc.
 */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */

/*
 * The threading model was changed in version 1.11.  Here's the details.
 *
 * We now use a dedicated "dispatcher" thread to process password messages.
 * The callback functions do not use synchronization guards when accessing
 * the password store, so those functions must only be called from the
 * dispatcher thread.
 *
 * Message processing behind a visible password dialog is the trickiest
 * part to get right.  We want to, at all costs, avoid repeatedly prompting
 * the user for the same password.  To that end, when the password dialog
 * for some key 'X' is visible, the dispatcher thread is blocked and
 * subsequent messages for key 'X' are expedited by redirecting them to a
 * special "express queue" [1].
 *
 * Once the password dialog is closed, all messages in the express queue
 * are either (depending on the user's response) cancelled or loaded with
 * the new password and returned to the sender.  The express queue is then
 * closed so that no new messages can enter, and the dispatcher thread
 * resumes its normal processing.
 *
 * [1] Yes, it's a stupid name, but it's the best I could think of.  It's
 *     somewhat analogous to an express lane on a freeway because messages
 *     get through faster.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <gtk/gtk.h>
#include <glib/gi18n-lib.h>

#include "e-passwords.h"
#include "libedataserver/e-flag.h"
#include "libedataserver/e-url.h"

#if WITH_GNOME_KEYRING
#include <gnome-keyring.h>
#endif

#define d(x)

typedef struct _EPassMsg EPassMsg;
typedef struct _EPassDialogData EPassDialogData;
typedef void (*EPassCallback) (EPassMsg *msg);

struct _EPassMsg {

	/* The 'expedite' flag indicates that the message
	 * is eligible for entry into the express queue. */

	/* dispatching */
	EPassCallback callback;
	gboolean expedite;
	EFlag *done;

	/* input/output */
	const gchar *component;
	const gchar *key;
	gchar *password;
	gpointer data;
};

struct _EPassDialogData {

	/* settings */
	GtkWindow *parent;
	const gchar *title;
	const gchar *prompt;
	gint flags;

	/* input/output */
	gboolean remember;
	gint response;

	/* synchronization */
	EFlag *done;
};

/* Forward Declarations */
static void ep_msg_dispatch (EPassMsg *msg);

static GThreadPool *dispatcher = NULL;
static GHashTable *password_cache = NULL;
static gboolean ep_online_state = TRUE;

/* The queue's lock also guards the key. */
static const gchar *express_key = NULL;
static GAsyncQueue *express_queue = NULL;

#define KEY_FILE_GROUP_PREFIX "Passwords-"
static GKeyFile *key_file = NULL;

static gchar *
ep_key_file_get_filename (void)
{
	/* XXX It would be nice to someday move this data elsewhere, or else
	 * fully migrate to GNOME Keyring or whatever software supercedes it.
	 * Evolution is one of the few remaining GNOME-2 applications that
	 * still uses the deprecated ~/.gnome2-private directory. */

	return g_build_filename (g_get_home_dir (),
		".gnome2_private", "Evolution", NULL);
}

static gchar *
ep_key_file_get_group (const gchar *component)
{
	return g_strconcat (KEY_FILE_GROUP_PREFIX, component, NULL);
}

static gchar *
ep_key_file_normalize_key (const gchar *key)
{
	/* XXX Previous code converted all slashes and equal signs in the
	 * key to underscores for use with "gnome-config" functions.  While
	 * it may not be necessary to convert slashes for use with GKeyFile,
	 * and an equal sign was most likely never encountered in a key, we
	 * continue to do the same for backward-compatibility. */

	gchar *normalized_key, *cp;

	normalized_key = g_strdup (key);
	for (cp = normalized_key; *cp != '\0'; cp++)
		if (*cp == '/' || *cp == '=')
			*cp = '_';

	return normalized_key;
}

static void
ep_key_file_load (void)
{
	gchar *filename;
	GError *error = NULL;

	filename = ep_key_file_get_filename ();

	if (!g_file_test (filename, G_FILE_TEST_EXISTS))
		goto exit;

	g_key_file_load_from_file (key_file, filename,
		G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS,
		&error);

	if (error != NULL) {
		g_warning ("%s", error->message);
		g_error_free (error);
	}

exit:
	g_free (filename);
}

static void
ep_key_file_save (void)
{
	gchar *contents;
	gchar *filename;
	gsize length;
	GError *error = NULL;

	filename = ep_key_file_get_filename ();
	contents = g_key_file_to_data (key_file, &length, NULL);

	g_file_set_contents (filename, contents, length, &error);

	if (error != NULL) {
		g_warning ("%s", error->message);
		g_error_free (error);
	}

	g_free (contents);
	g_free (filename);
}

static gchar *
ep_password_encode (const gchar *password)
{
	/* XXX The previous Base64 encoding function did not encode the
	 * password's trailing nul byte.  This makes decoding the Base64
	 * string into a nul-terminated password more difficult, but we
	 * continue to do it this way for backward-compatibility. */

	gsize length = strlen (password);
	return g_base64_encode ((const guchar *) password, length);
}

static gchar *
ep_password_decode (const gchar *encoded_password)
{
	/* XXX The previous Base64 encoding function did not encode the
	 * password's trailing nul byte, so we have to append a nul byte
	 * to the decoded data to make it a nul-terminated string. */

	gchar *password;
	gsize length;

	password = (gchar *) g_base64_decode (encoded_password, &length);
	password = g_realloc (password, length + 1);
	password[length] = '\0';

	return password;
}

static gpointer
ep_init (gpointer data)
{
	/* Initialize the data structures with unbounded lifetimes. */

	dispatcher = g_thread_pool_new (
		(GFunc) ep_msg_dispatch, NULL, 1, FALSE, NULL);

	password_cache = g_hash_table_new_full (
		g_str_hash, g_str_equal, g_free, g_free);

	express_queue = g_async_queue_new ();

#ifdef WITH_GNOME_KEYRING
	if (!gnome_keyring_is_available ()) {
		key_file = g_key_file_new ();
		ep_key_file_load ();
	}
#else
	key_file = g_key_file_new ();
	ep_key_file_load ();
#endif

	return NULL;
}

static EPassMsg *
ep_msg_new (EPassCallback callback, gboolean expedite)
{
	static GOnce once = G_ONCE_INIT;
	EPassMsg *msg;

	g_once (&once, ep_init, NULL);

	msg = g_slice_new0 (EPassMsg);
	msg->callback = callback;
	msg->expedite = expedite;
	msg->done = e_flag_new ();

	return msg;
}

static void
ep_msg_dispatch (EPassMsg *msg)
{
	msg->callback (msg);
	e_flag_set (msg->done);
}

static gboolean
ep_msg_expedite (EPassMsg *msg)
{
	gboolean success = FALSE;

	g_async_queue_lock (express_queue);
	if (express_key != NULL && strcmp (msg->key, express_key) == 0) {
		g_async_queue_push_unlocked (express_queue, msg);
		success = TRUE;
	}
	g_async_queue_unlock (express_queue);

	return success;
}

static void
ep_msg_send (EPassMsg *msg)
{
	GMainContext *context;

	context = g_main_context_default (); 

	/* Expedite flag requires a key. */
	g_assert (!msg->expedite || msg->key != NULL);

	if (!(msg->expedite && ep_msg_expedite (msg)))
		g_thread_pool_push (dispatcher, msg, NULL);

	/* Don't block the main loop. */
	if (g_main_context_is_owner (context))
		while (!e_flag_is_set (msg->done))
			g_main_context_iteration (context, TRUE);
	else
		e_flag_wait (msg->done);
}

static void
ep_msg_free (EPassMsg *msg)
{
	e_flag_free (msg->done);
	g_slice_free (EPassMsg, msg);
}

static EPassDialogData *
ep_dialog_data_new (GtkWindow *parent, const gchar *title,
                    const gchar *prompt, gboolean remember, gint flags)
{
	EPassDialogData *data;

	data = g_slice_new0 (EPassDialogData);
	data->done = e_flag_new ();

	data->parent = parent;
	data->title = title;
	data->prompt = prompt;
	data->remember = remember;
	data->flags = flags;

	return data;
}

static void
ep_dialog_data_free (EPassDialogData *data)
{
	e_flag_free (data->done);
	g_slice_free (EPassDialogData, data);
}

static gboolean
ep_dialog_run (EPassMsg *msg)
{
	/* This function must run in the main thread. */

	EPassDialogData *data = msg->data;
	GtkBox *box;
	GtkWidget *dialog;
	GtkWidget *entry;
	GtkWidget *check_button = NULL;
	gboolean visible;
	gint remember;

	visible = !(data->flags & E_PASSWORDS_SECRET);
	remember = data->flags & E_PASSWORDS_REMEMBER_MASK;

	/* Create a standard message dialog. */
	dialog = gtk_message_dialog_new (
		data->parent, 0, GTK_MESSAGE_QUESTION,
		GTK_BUTTONS_OK_CANCEL, "%s", data->prompt);
	gtk_window_set_title (GTK_WINDOW (dialog), data->title);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);

	box = GTK_BOX (GTK_DIALOG (dialog)->vbox);

	/* Add a GtkEntry for the password. */
	entry = gtk_entry_new ();
	gtk_box_pack_start (box, entry, TRUE, FALSE, 3);
	gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
	gtk_entry_set_visibility (GTK_ENTRY (entry), visible);
	if (msg->password != NULL) {
		gtk_entry_set_text (GTK_ENTRY (entry), msg->password);
		g_free (msg->password);
		msg->password = NULL;
	}
	atk_object_set_description (
		gtk_widget_get_accessible (entry), data->prompt);
	gtk_widget_grab_focus (entry);
	gtk_widget_show (entry);

	/* Add a GtkCheckButton for the "remember" option. */
	if (remember != E_PASSWORDS_REMEMBER_NEVER) {
		if (data->flags & E_PASSWORDS_PASSPHRASE)
			check_button = gtk_check_button_new_with_mnemonic (
				(remember == E_PASSWORDS_REMEMBER_FOREVER) ?
				_("_Remember this passphrase") :
				_("_Remember this passphrase for the "
				"remainder of this session"));
		else
			check_button = gtk_check_button_new_with_mnemonic (
				(remember == E_PASSWORDS_REMEMBER_FOREVER) ?
				_("_Remember this password") :
				_("_Remember this password for the "
				"remainder of this session"));
		gtk_box_pack_start (box, check_button, TRUE, FALSE, 3);
		gtk_toggle_button_set_active (
			GTK_TOGGLE_BUTTON (check_button), data->remember);
		if (data->flags & E_PASSWORDS_DISABLE_REMEMBER)
			gtk_widget_set_sensitive (check_button, FALSE);
		gtk_widget_show (check_button);
	}

	/* Run the dialog and collect the results. */
	data->response = gtk_dialog_run (GTK_DIALOG (dialog));
	msg->password = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry)));
	if (check_button != NULL)
		data->remember = gtk_toggle_button_get_active (
			GTK_TOGGLE_BUTTON (check_button));
	else
		data->remember = FALSE;

	gtk_widget_destroy (dialog);
	e_flag_set (data->done);

	return FALSE;
}

/* The rest of these static functions must run in the dispatcher thread. */

#if WITH_GNOME_KEYRING
static void
ep_clear_passwords_keyring (EPassMsg *msg)
{
	GnomeKeyringAttributeList *attributes;
	GnomeKeyringAttribute attribute;
	GnomeKeyringResult result;
	GList *matches = NULL, *tmp;	
	char *path;
	char *default_keyring = NULL;

	result = gnome_keyring_get_default_keyring_sync (&default_keyring);
	if (!default_keyring) {
	        if (gnome_keyring_create_sync ("default", NULL) != GNOME_KEYRING_RESULT_OK)
			return;
	        default_keyring = g_strdup ("default");			
	}

	d(g_print("Get Default %d\n", result));
	
	/* Not called at all */
	path = g_strdup ("Evolution");
	attributes = gnome_keyring_attribute_list_new ();

	attribute.name = g_strdup ("application");
	attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
	attribute.value.string = path;
	g_array_append_val (attributes, attribute);

	result = gnome_keyring_find_items_sync (GNOME_KEYRING_ITEM_NETWORK_PASSWORD, attributes, &matches);
	d(g_print ("Find Items %d\n", result));
		
	gnome_keyring_attribute_list_free (attributes);

	if (result) {
		g_print ("Couldn't clear password");
	} else {
		for (tmp = matches; tmp; tmp = tmp->next) {
			result = gnome_keyring_item_delete_sync (default_keyring, ((GnomeKeyringFound *) tmp->data)->item_id);
			d(g_print("Delete Items %d\n", result));
		}
	}
	
	g_free (default_keyring);
}
#endif

static void
ep_clear_passwords_keyfile (EPassMsg *msg)
{
	gchar *group;
	GError *error = NULL;

	group = ep_key_file_get_group (msg->component);

	g_key_file_remove_group (key_file, group, &error);
	if (error == NULL)
		ep_key_file_save ();
	else {
		g_warning ("%s", error->message);
		g_error_free (error);
	}

	g_free (group);
}

static void
ep_clear_passwords (EPassMsg *msg)
{
#if WITH_GNOME_KEYRING
	if (gnome_keyring_is_available ())
		ep_clear_passwords_keyring (msg);
	else
		ep_clear_passwords_keyfile (msg);
#else
	ep_clear_passwords_keyfile (msg);
#endif
}

#if WITH_GNOME_KEYRING
static void
ep_forget_passwords_keyring (EPassMsg *msg)
{
	GnomeKeyringAttributeList *attributes;
	GnomeKeyringAttribute attribute;
	GnomeKeyringResult result;
	GList *matches = NULL, *tmp;	
	char *path;
	char *default_keyring = NULL;

	result = gnome_keyring_get_default_keyring_sync (&default_keyring);
	if (!default_keyring) {
	        if (gnome_keyring_create_sync ("default", NULL) != GNOME_KEYRING_RESULT_OK)
			return;
	        default_keyring = g_strdup ("default");			
	}	
	d(g_print("Get Default %d\n", result));
	
	path = g_strdup ("Evolution");
	attributes = gnome_keyring_attribute_list_new ();

	attribute.name = g_strdup ("application");
	attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
	attribute.value.string = path;
	g_array_append_val (attributes, attribute);


	result = gnome_keyring_find_items_sync (GNOME_KEYRING_ITEM_NETWORK_PASSWORD, attributes, &matches);
	d(g_print ("Find Items %d\n", result));
		
	gnome_keyring_attribute_list_free (attributes);

	if (result) {
		g_print ("Couldn't clear password");
	} else {
		for (tmp = matches; tmp; tmp = tmp->next) {
			result = gnome_keyring_item_delete_sync (default_keyring, ((GnomeKeyringFound *) tmp->data)->item_id);
			d(g_print("Delete Items %d\n", result));
		}
	}
	
	g_free (default_keyring);

	/* free up the session passwords */
	g_hash_table_remove_all (password_cache);
}
#endif

static void
ep_forget_passwords_keyfile (EPassMsg *msg)
{
	gchar **groups;
	gsize length, ii;

	g_hash_table_remove_all (password_cache);

	groups = g_key_file_get_groups (key_file, &length);
	for (ii = 0; ii < length; ii++) {
		if (!g_str_has_prefix (groups[ii], KEY_FILE_GROUP_PREFIX))
			continue;
		g_key_file_remove_group (key_file, groups[ii], NULL);
	}
	ep_key_file_save ();
	g_strfreev (groups);
}

static void
ep_forget_passwords (EPassMsg *msg)
{
#if WITH_GNOME_KEYRING
	if (gnome_keyring_is_available ())
		ep_forget_passwords_keyring (msg);
	else
		ep_forget_passwords_keyfile (msg);
#else
	ep_forget_passwords_keyfile (msg);
#endif
}

#if WITH_GNOME_KEYRING
static void
ep_remember_password_keyring (EPassMsg *msg)
{
	gpointer value;

	value = g_hash_table_lookup (password_cache, msg->key);
	if (value != NULL) {
		/* add it to the on-disk cache of passwords */
		GnomeKeyringAttributeList *attributes;
		GnomeKeyringAttribute attribute;
		GnomeKeyringResult result;
		EUri *uri = e_uri_new (msg->key);
		guint32 item_id;

		if (!strcmp (uri->protocol, "ldap") && !uri->user) {
			/* LDAP doesnt use username in url. Let the url be the user key. So safe it */
			char *keycopy = g_strdup (msg->key);
			int i;
			
			for (i = 0; i < strlen (keycopy); i ++)
				if (keycopy[i] == '/' || keycopy[i] =='=')
					keycopy[i] = '_';		
			uri->user = keycopy;
		}
		
		attributes = gnome_keyring_attribute_list_new ();

		attribute.name = g_strdup ("user");
		attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
		attribute.value.string = g_strdup (uri->user);
		g_array_append_val (attributes, attribute);
	
		attribute.name = g_strdup ("server");
		attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
		attribute.value.string = g_strdup (uri->host);
		g_array_append_val (attributes, attribute);
	 
		attribute.name = g_strdup ("application");
		attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
		attribute.value.string = g_strdup ("Evolution");
		g_array_append_val (attributes, attribute);		
		
		result = gnome_keyring_item_create_sync (NULL, /* Use default keyring */
						         GNOME_KEYRING_ITEM_NETWORK_PASSWORD, /* type */
				   			 msg->key, /* name */
				   			 attributes, /* attribute list */
				   			 value, /* password */
				   			 TRUE, /* Update if already exists */
				   			 &item_id);
	
		d(g_print("Remember %s: %d/%d\n", msg->key, result, item_id));
		/* now remove it from our session hash */
		g_hash_table_remove (password_cache, msg->key);
		
		e_uri_free (uri);
	}
}
#endif

static void
ep_remember_password_keyfile (EPassMsg *msg)
{
	gchar *group, *key, *password;

	password = g_hash_table_lookup (password_cache, msg->key);
	if (password == NULL) {
		g_warning ("Password for key \"%s\" not found", msg->key);
		return;
	}

	group = ep_key_file_get_group (msg->component);
	key = ep_key_file_normalize_key (msg->key);
	password = ep_password_encode (password);

	g_hash_table_remove (password_cache, msg->key);
	g_key_file_set_string (key_file, group, key, password);
	ep_key_file_save ();

	g_free (group);
	g_free (key);
	g_free (password);
}

static void
ep_remember_password (EPassMsg *msg)
{
#if WITH_GNOME_KEYRING
	if (gnome_keyring_is_available ())
		ep_remember_password_keyring (msg);
	else
		ep_remember_password_keyfile (msg);
#else
	ep_remember_password_keyfile (msg);
#endif
}

#if WITH_GNOME_KEYRING
static void
ep_forget_password_keyring (EPassMsg *msg)
{
	GnomeKeyringAttributeList *attributes;
	GnomeKeyringAttribute attribute;
	GnomeKeyringResult result;
	GList *matches = NULL, *tmp;	
	char *default_keyring = NULL;	
	EUri *uri = e_uri_new (msg->key);

	if (!strcmp (uri->protocol, "ldap") && !uri->user) {
		/* LDAP doesnt use username in url. Let the url be the user key. So safe it */
		char *keycopy = g_strdup (msg->key);
		int i;

		for (i = 0; i < strlen (keycopy); i ++)
			if (keycopy[i] == '/' || keycopy[i] =='=')
				keycopy[i] = '_';		
		uri->user = keycopy;
	}
	    
	g_hash_table_remove (password_cache, msg->key);

	if (!uri->host && !uri->user) {
		/* No need to remove from keyring for pass phrases */
		return;
	}
	
	result = gnome_keyring_get_default_keyring_sync (&default_keyring);
	if (!default_keyring) {
	        if (gnome_keyring_create_sync ("default", NULL) != GNOME_KEYRING_RESULT_OK)
			return;
	        default_keyring = g_strdup ("default");			
	}

	d(g_print("Get Default %d\n", result));
	
	attributes = gnome_keyring_attribute_list_new ();
	attribute.name = g_strdup ("user");
	attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
	attribute.value.string = g_strdup(uri->user);;
	g_array_append_val (attributes, attribute);

	attributes = gnome_keyring_attribute_list_new ();
	attribute.name = g_strdup ("server");
	attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
	attribute.value.string = g_strdup(uri->host);;
	g_array_append_val (attributes, attribute);

	attributes = gnome_keyring_attribute_list_new ();
	attribute.name = g_strdup ("application");
	attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
	attribute.value.string = g_strdup("Evolution");;
	g_array_append_val (attributes, attribute);	

	result = gnome_keyring_find_items_sync (GNOME_KEYRING_ITEM_NETWORK_PASSWORD, attributes, &matches);
	d(g_print ("Find Items %d\n", result));
		
	gnome_keyring_attribute_list_free (attributes);

	if (result) {
		g_print ("Couldn't clear password");
	} else {
		for (tmp = matches; tmp; tmp = tmp->next) {
			GArray *pattr = ((GnomeKeyringFound *) tmp->data)->attributes;
			int i;
			GnomeKeyringAttribute *attr;
			gboolean accept = TRUE;
			guint present = 0;

			for (i =0; (i < pattr->len) && accept; i++)
			{
				attr = &g_array_index (pattr, GnomeKeyringAttribute, i);
				if (!strcmp(attr->name, "user")) {
					present++;
					if (strcmp (attr->value.string, uri->user))
						accept = FALSE;
				} else if (!strcmp(attr->name, "server")) {
					present++;
					if (strcmp (attr->value.string, uri->host))
						accept = FALSE;						
				}
			}
				if (present == 2 && accept) {
					result = gnome_keyring_item_delete_sync (default_keyring, ((GnomeKeyringFound *) tmp->data)->item_id);
					d(g_print("Delete Items %s %s %d\n", uri->host, uri->user, result));			
				}
		}	

	}
	
	g_free (default_keyring);
}
#endif

static void
ep_forget_password_keyfile (EPassMsg *msg)
{
	gchar *group, *key;
	GError *error = NULL;

	g_hash_table_remove (password_cache, msg->key);

	group = ep_key_file_get_group (msg->component);
	key = ep_key_file_normalize_key (msg->key);

	g_key_file_remove_key (key_file, group, key, &error);
	if (error == NULL)
		ep_key_file_save ();
	else {
		g_warning ("%s", error->message);
		g_error_free (error);
	}

	g_free (group);
	g_free (key);
}

static void
ep_forget_password (EPassMsg *msg)
{
#if WITH_GNOME_KEYRING
	if (gnome_keyring_is_available ())
		ep_forget_password_keyring (msg);
	else
		ep_forget_password_keyfile (msg);
#else
	ep_forget_password_keyfile (msg);
#endif
}

#if WITH_GNOME_KEYRING
static void
ep_get_password_keyring (EPassMsg *msg)
{
	char *passwd;
	GnomeKeyringAttributeList *attributes;
	GnomeKeyringAttribute attribute;
	GnomeKeyringResult result;
	GList *matches = NULL, *tmp;

	passwd = g_hash_table_lookup (password_cache, msg->key);
	if (passwd) {
		msg->password = g_strdup (passwd);
	} else {
		EUri *uri = e_uri_new (msg->key);
		
		if (!strcmp (uri->protocol, "ldap") && !uri->user) {
			/* LDAP doesnt use username in url. Let the url be the user key. So safe it */
			char *keycopy = g_strdup (msg->key);
			int i;

			for (i = 0; i < strlen (keycopy); i ++)
				if (keycopy[i] == '/' || keycopy[i] =='=')
					keycopy[i] = '_';		
			uri->user = keycopy;
		}
		
		if (uri->host &&  uri->user) {
			/* We dont store passphrases.*/

			attributes = gnome_keyring_attribute_list_new ();
			attribute.name = g_strdup ("user");
			attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
			attribute.value.string = g_strdup(uri->user);;
			g_array_append_val (attributes, attribute);
			printf("get %s %s\n", attribute.value.string, msg->key);

			attributes = gnome_keyring_attribute_list_new ();
			attribute.name = g_strdup ("server");
			attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
			attribute.value.string = g_strdup(uri->host);;
			g_array_append_val (attributes, attribute);
		
			attributes = gnome_keyring_attribute_list_new ();
			attribute.name = g_strdup ("application");
			attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
			attribute.value.string = g_strdup("Evolution");
			g_array_append_val (attributes, attribute);	

			result = gnome_keyring_find_items_sync (GNOME_KEYRING_ITEM_NETWORK_PASSWORD, attributes, &matches);
			d(g_print ("Find Items %d\n", result));
			
			gnome_keyring_attribute_list_free (attributes);

			if (result) {
				g_print ("Couldn't Get password %d\n", result);
			} else {
				/* FIXME: What to do if this returns more than one? */
				for (tmp = matches; tmp; tmp = tmp->next) {
					GArray *pattr = ((GnomeKeyringFound *) tmp->data)->attributes;
					int i;
					GnomeKeyringAttribute *attr;
					gboolean accept = TRUE;
					guint present = 0;

					for (i =0; (i < pattr->len) && accept; i++)
					{
						attr = &g_array_index (pattr, GnomeKeyringAttribute, i);

						if (!strcmp(attr->name, "user")) {
							present++;
							if (strcmp (attr->value.string, uri->user))
								accept = FALSE;
						} else if (!strcmp(attr->name, "server")) {
							present++;
							if (strcmp (attr->value.string, uri->host))
								accept = FALSE;						
						}
					}
					if (present == 2 && accept)
						msg->password = g_strdup (((GnomeKeyringFound *) tmp->data)->secret);
				}	
			}
		}
		
	}
}	
#endif

static void
ep_get_password_keyfile (EPassMsg *msg)
{	
	gchar *group, *key, *password;
	GError *error = NULL;

	password = g_hash_table_lookup (password_cache, msg->key);
	if (password != NULL) {
		msg->password = g_strdup (password);
		return;
	}

	group = ep_key_file_get_group (msg->component);
	key = ep_key_file_normalize_key (msg->key);

	password = g_key_file_get_string (key_file, group, key, &error);
	if (password != NULL) {
		msg->password = ep_password_decode (password);
		g_free (password);
	} else {
		g_warning ("%s", error->message);
		g_error_free (error);
	}

	g_free (group);
	g_free (key);
}

static void
ep_get_password (EPassMsg *msg)
{
#if WITH_GNOME_KEYRING
	if (gnome_keyring_is_available ())
		ep_get_password_keyring (msg);
	else
		ep_get_password_keyfile (msg);
#else
	ep_get_password_keyfile (msg);
#endif
}

static void
ep_add_password (EPassMsg *msg)
{
	gchar *key = g_strdup (msg->key);
	gchar *password = g_strdup (msg->password);

	g_hash_table_insert (password_cache, key, password);
}

static void
ep_ask_password (EPassMsg *msg)
{
	EPassDialogData *data = msg->data;
	gchar *password;
	gint remember;

	/* Open the express queue for this key.  This allows us to catch
	 * incoming operations on this key while waiting for the user to
	 * provide a password.  Once we have the password, immediately
	 * process any messages in the express queue. */
	g_async_queue_lock (express_queue);
	express_key = msg->key;
	g_async_queue_unlock (express_queue);

	if (data->flags & E_PASSWORDS_REPROMPT)
		ep_get_password (msg);

	/* Run the password dialog in the main thread. */
	g_idle_add ((GSourceFunc) ep_dialog_run, msg);

	/* Wait for the user to respond. */
	e_flag_wait (data->done);

	/* If the user cancelled the dialog then we need to dispatch
	 * any messages that we trapped in the express queue. */
	if (data->response != GTK_RESPONSE_OK) {
		g_async_queue_lock (express_queue);
		express_key = NULL;
		while ((msg = g_async_queue_try_pop_unlocked (express_queue)))
			ep_msg_dispatch (msg);
		g_async_queue_unlock (express_queue);
		return;
	}

	remember = data->flags & E_PASSWORDS_REMEMBER_MASK;

	if (remember == E_PASSWORDS_REMEMBER_NEVER)
		goto cleanup;

	if (data->remember || remember == E_PASSWORDS_REMEMBER_FOREVER)
		ep_add_password (msg);

	if (data->remember && remember == E_PASSWORDS_REMEMBER_FOREVER)
		ep_remember_password (msg);

cleanup:

	password = g_strdup (msg->password);

	/* Close the express queue and process any messages there. */
	g_async_queue_lock (express_queue);
	express_key = NULL;
	while ((msg = g_async_queue_try_pop_unlocked (express_queue))) {
		msg->password = g_strdup (password);
		e_flag_set (msg->done);
	}
	g_async_queue_unlock (express_queue);

	g_free (password);
}

/**
 * e_passwords_init:
 *
 * This function is here for backward-compatibility.  It does nothing.
 **/
void
e_passwords_init (void)
{
}

/**
 * e_passwords_shutdown:
 *
 * This function is here for backward-compatibility.  It does nothing.
 **/
void
e_passwords_shutdown (void)
{
}

/**
 * e_passwords_cancel:
 * 
 * This function is here for backward-compatibility.  It does nothing.
 **/
void
e_passwords_cancel(void)
{
}

/**
 * e_passwords_set_online:
 * @online: 
 * 
 * Set the offline-state of the application.  This is a work-around
 * for having the backends fully offline aware, and returns a
 * cancellation response instead of prompting for passwords.
 *
 * FIXME: This is not a permanent api, review post 2.0.
 **/
void
e_passwords_set_online (gboolean online)
{
	/* TODO We could check that a request is open and close it,
	 *      or maybe who cares */
	ep_online_state = online;
}

/**
 * e_passwords_forget_passwords:
 *
 * Forgets all cached passwords, in memory and on disk.
 **/
void
e_passwords_forget_passwords (void)
{
	EPassMsg *msg;

	msg = ep_msg_new (ep_forget_passwords, FALSE);

	ep_msg_send (msg);
	ep_msg_free (msg);
}

/**
 * e_passwords_clear_passwords:
 * @component: component name of the caller
 *
 * Forgets all disk cached passwords for the component.
 **/
void
e_passwords_clear_passwords (const gchar *component)
{
	EPassMsg *msg;

	g_return_if_fail (component != NULL);

	msg = ep_msg_new (ep_clear_passwords, FALSE);
	msg->component = component;

	ep_msg_send (msg);
	ep_msg_free (msg);
}

/**
 * e_passwords_remember_password:
 * @component: component name of the caller
 * @key: the key
 *
 * Saves the password associated with @key to disk.
 **/
void
e_passwords_remember_password (const gchar *component, const gchar *key)
{
	EPassMsg *msg;

	g_return_if_fail (component != NULL);
	g_return_if_fail (key != NULL);

	msg = ep_msg_new (ep_remember_password, FALSE);
	msg->component = component;
	msg->key = key;

	ep_msg_send (msg);
	ep_msg_free (msg);
}

/**
 * e_passwords_forget_password:
 * @component: component name of the caller
 * @key: the key
 *
 * Forgets the password associated with @key, in memory and on disk.
 **/
void
e_passwords_forget_password (const gchar *component, const gchar *key)
{
	EPassMsg *msg;

	g_return_if_fail (component != NULL);
	g_return_if_fail (key != NULL);

	msg = ep_msg_new (ep_forget_password, TRUE);
	msg->component = component;
	msg->key = key;

	ep_msg_send (msg);
	ep_msg_free (msg);
}

/**
 * e_passwords_get_password:
 * @component: component name of the caller
 * @key: the key
 *
 * Return value: the password associated with @key, or %NULL.  Caller
 * must free the returned password.
 **/
char *
e_passwords_get_password (const gchar *component, const gchar *key)
{
	EPassMsg *msg;
	char *password;

	g_return_val_if_fail (component != NULL, NULL);
	g_return_val_if_fail (key != NULL, NULL);

	msg = ep_msg_new (ep_get_password, TRUE);
	msg->component = component;
	msg->key = key;

	ep_msg_send (msg);
	password = msg->password;
	ep_msg_free (msg);

	return password;
}

/**
 * e_passwords_add_password:
 * @key: a key
 * @password: the password for @key
 *
 * This stores the @key/@passwd pair in the current session's password
 * hash.
 **/
void
e_passwords_add_password (const gchar *key, const gchar *password)
{
	EPassMsg *msg;

	g_return_if_fail (key != NULL);
	g_return_if_fail (password != NULL);

	msg = ep_msg_new (ep_add_password, FALSE);
	msg->key = key;
	msg->password = g_strdup (password);

	ep_msg_send (msg);
	g_free (msg->password);
	ep_msg_free (msg);
}

/**
 * e_passwords_ask_password:
 * @title: title for the password dialog
 * @component: component name of the caller
 * @key: key to store the password under
 * @prompt: prompt string
 * @secret: whether or not the password text should be ***ed out
 * @remember_type: whether or not to offer to remember the password,
 * and for how long.
 * @remember: on input, the default state of the remember checkbox.
 * on output, the state of the checkbox when the dialog was closed.
 * @parent: parent window of the dialog, or %NULL
 *
 * Asks the user for a password.
 *
 * Return value: the password, which the caller must free, or %NULL if
 * the user cancelled the operation. *@remember will be set if the
 * return value is non-%NULL and @remember_type is not
 * E_PASSWORDS_DO_NOT_REMEMBER.
 **/
gchar *
e_passwords_ask_password (const gchar *title,
			  const gchar *component,
			  const gchar *key,
			  const gchar *prompt,
			  EPasswordsRememberType type,
			  gboolean *p_remember,
			  GtkWindow *parent)
{
	EPassMsg *msg;
	EPassDialogData *data;
	gboolean remember;
	gchar *password;

	g_return_val_if_fail (title != NULL, NULL);
	g_return_val_if_fail (component != NULL, NULL);
	g_return_val_if_fail (prompt != NULL, NULL);

	if ((type & E_PASSWORDS_ONLINE) && !ep_online_state)
		return NULL;

	remember = (p_remember != NULL) ? *p_remember : FALSE;
	data = ep_dialog_data_new (parent, title, prompt, remember, type);

	msg = ep_msg_new (ep_ask_password, TRUE);
	msg->component = component;
	msg->key = (key != NULL) ? key : "";
	msg->data = data;

	ep_msg_send (msg);
	password = msg->password;
	remember = data->remember;
	ep_msg_free (msg);

	ep_dialog_data_free (data);

	if (p_remember != NULL)
		*p_remember = remember;

	return password;
}
